/* 
 * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
 *
 * for details see file COPYING in the DOSEMU distribution
 */

/* very simple semaphores */

#ifndef MUTEX_H
#define MUTEX_H

typedef int sem_t;

#endif /* MUTEX_H */
